#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-28 13:31:51
# Filename      : command.py
# Description   : 

from __future__ import print_function, unicode_literals
import optparse

def parse():
    parser = optparse.OptionParser()
    parser.add_option('-l', '--list', dest = 'list_nodes', 
            help = 'list all proxy nodes', action = 'store_true')

    options, args = parser.parse_args()
    options.args = args

    return options




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 11:13:35
# Filename      : api.py
# Description   : 
from __future__ import print_function, unicode_literals
import requests
from dsb_requests import logs
from dsb_requests import patch_const
reload(patch_const)
import random
import time
import socket

def _request(method, url, data = None):
    try:
        response = requests.request(method, 
                'http://{hostname}:{port}{url}'.format(
                    hostname = API_HOSTNAME, port = API_PORT, url = url), 
                headers = {
                    'dsb_client_key'    :       DSB_CLIENT_KEY,
                    }, 
                data = data)

        if response.json()['status_code'] != 0:
            logs.print_error(response.json()['status_msg'] or 'error')
            return

        return response

    except requests.exceptions.RequestException:
        return None


def fetch_proxy_node_response():
    """获取所有可用的节点信息"""
    response = _request('GET', '/api/proxy/node')
    if not response:
        return []

    nodes = response.json()['nodes']
    return nodes

def feedback_node(node_ip, username, password):
    _request('PUT', '/api/proxy/node', data = {
        'node_ip'       :       node_ip,
        'username'      :       username,
        'password'      :       password,
        })




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-21 09:05:56
# Filename      : _init.py
# Description   : 
from __future__ import print_function, unicode_literals
import os
import json

__ALL__ = ['init_dsb_requests']

CFG_PATH = os.path.join(os.path.expandvars('$HOME'), '.dsb_requestsrc')

def read(prompt, default = None):
    while True:
        line = raw_input(prompt).strip()
        if line:
            return line

        elif default is not None:
            return default
        

def init_dsb_requests():
    api_hostname = read('please enter api hostname:')
    api_port = int(read('please enter api port(default 16227):', 16227))
    dsb_client_key = read('please enter dsb client key:')
    with open(CFG_PATH, 'wb') as fd:
        fd.write(json.dumps({
            'API_HOSTNAME'      :       api_hostname,
            'API_PORT'          :       api_port,
            'DSB_CLIENT_KEY'    :       dsb_client_key,
            }))

def read_dsb_requests_init_data():
    if not os.path.exists(CFG_PATH):
        return {}

    with open(CFG_PATH, 'rb') as fd:
        return json.loads(fd.read())




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 11:19:15
# Filename      : logs.py
# Description   : 
from __future__ import print_function, unicode_literals
import logging
import sys

def print_error(error):
    logging.error(error)

def alert_error(error):
    print_error(error)
    sys.exit(-1)

def debug(msg):
    logging.debug(msg)




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-29 14:09:32
# Filename      : _adapters.py
# Description   : 
from __future__ import print_function, unicode_literals
from requests.adapters import HTTPAdapter
from dsb_requests import _api
from dsb_requests import tools
import socket

class UDPHTTPAdapter(HTTPAdapter):
    def send(self, request, timeout = None, proxies = None, **kwargs):
        proxy_node = dict(proxies)
        conn = self.get_conn_to_proxy_node(proxy_node, timeout or 1) # 进行udp打洞
        assert proxy_node, 'not has proxy_node'
        assert not proxy_node['internet']
        request.url = request.url[3:] # 删除url前的udp

    def get_conn_to_proxy_node(self, proxy_node, timeout):
        conn = tools.made_conn()
        conn.bind(('', 0))
        node_port_in_lan = _api.notify_proxy_node_access_me(proxy_node, conn)
        print('node_port_in_lan', node_port_in_lan)




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-19 14:25:22
# Filename      : _proxy.py
# Description   : 
from __future__ import print_function, unicode_literals
import requests
from dsb_requests import patch_const
reload(patch_const)
from dsb_requests import _api
import socket
import time
import threading

class ProxyManager(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            _instance = object.__new__(cls, *args, **kwargs)
            cls.__init__(_instance, *args, **kwargs)
            cls._instance = _instance

        return cls._instance

    def __init__(self):
        self._last_node = None
        self._node_use_counter = 0
        self.max_node_use = int(NODE_CACHE_COUNTER)
        self._rlock = threading.RLock()

    @property
    def node_use_counter(self):
        if not self._last_node:
            self._node_use_counter = 0

        return self._node_use_counter

    def get_cache_node(self):
        if not self._last_node:
            return None

        if self.node_use_counter > int(NODE_CACHE_COUNTER):
            return None

        return self._last_node

    @property
    def last_node(self):
        return self._last_node

    @last_node.setter
    def last_node(self, node):
        self._last_node = node
        self._node_use_counter = 0

    def feedback_node(self, node):
        """把这个节点告诉服务器，需要更换了"""
        _api.feedback_node(node['ip'], node['username'], node['password'])
        self.flush_cache()

    def incr_use_counter(self):
        self._node_use_counter += 1

    def flush_cache(self):
        self._last_node = None
        self._node_use_counter = 0

    def get_best_proxy_node(self):
        with self._rlock:
            return self._get_best_proxy_node()

    def _get_best_proxy_node(self):
        cache_node = self.get_cache_node()
        if cache_node:
            return cache_node

        nodes = _api.fetch_proxy_node_response()
        if not nodes:
            return None

        alive_nodes = []
        for node in nodes:
            try:
                node['delay'] = self.get_node_delay(node)
                alive_nodes.append(node)

            except socket.error as ex:
                continue

        if not alive_nodes:
            return 

        best_node = min(alive_nodes, key = lambda node: node['delay'])

        self.last_node = best_node
        return best_node

    def get_node_delay(self, node):
        begin_time = time.time()
        test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        test_socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

        test_socket.settimeout(1)
        test_socket.connect((node['ip'], node['port']))
        delay = time.time() - begin_time
        test_socket.close()

        return delay




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 21:06:48
# Filename      : patch_const.py
# Description   : 
from __future__ import print_function, unicode_literals
from dsb_requests import _init
import inspect
import os

DEFAULT_OPTIONS = {
        'PROXY_RETRY'           :       '2',
        'REQUEST_RETRY'         :       '2',
        'REQUEST_RETRY_INTERVAL':       '0',
        'NODE_CACHE_COUNTER'    :       '10',
        'DSB_CLIENT_KEY'        :       '',
        }

def get_options():
    import os
    _options = DEFAULT_OPTIONS.copy()
    _options.update(_init.read_dsb_requests_init_data())
    for option_name in _options.keys():
        if option_name in os.environ:
            _options[option_name] = os.environ[option_name]

    return _options

if __name__ != "__main__":
    for context in inspect.stack()[1:]:
        frame = context[0]
        if frame.f_code.co_name == '<module>':
            frame.f_globals.update(get_options())
            break




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 20:00:48
# Filename      : tools.py
# Description   : 
from __future__ import print_function, unicode_literals
import threading
from functools import wraps

def is_string(s):
    return isinstance(s, (str, unicode))

def is_list_or_tuple(data):
    return isinstance(data, (tuple, list))

def to_list(data):
    return data if is_list_or_tuple(data) else [data]

def async(func):
    @wraps(func)
    def innerwrap(*args, **kwargs):
        t = threading.Thread(target = func, args = args, kwargs = kwargs)
        t.setDaemon(True)
        t.start()

    return innerwrap




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 17:57:21
# Filename      : setup.py
# Description   : 
from setuptools import setup

setup(
        name = 'dsb_requests',
        version = '0.9.7',
        author = 'tuxpy',
        include_package_data = True,
        packages = [
            'dsb_requests',
            ],
        install_requires = ['requests', 'lxml', 'mock'],
        description = 'duoshoubang \'s requests',
        scripts = ['bin/dsb_requests'],
        )




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 11:10:13
# Filename      : requests.py
# Description   : 
from __future__ import print_function, unicode_literals
import requests
from dsb_requests import patch_const
reload(patch_const)
from dsb_requests import _proxy
from dsb_requests import tools
from functools import partial
import mock
import lxml.etree
import time
import re
import urlparse
import threading

if hasattr(requests, 'packages'):
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = 'ALL'

    requests.packages.urllib3.disable_warnings() # see https://urllib3.readthedocs.org/en/latest/security.html#disabling-warnings

_proxy_manager = _proxy.ProxyManager()
_flush_lock = threading.RLock()

def flush_ip(async = True):
    last_node = _proxy_manager.last_node
    if not last_node:
        return

    if async:
        tools.async(_proxy_manager.feedback_node)(last_node)

    else:
        with _flush_lock:
            _proxy_manager.feedback_node(last_node)

class HTMLSelector(object):
    def __init__(self, response):
        self.response = response
        self._etree = None

    @property
    def etree(self):
        if self._etree is None:
            self._etree = lxml.etree.HTML(self.response.text.strip())

        return self._etree

    def get(self, xpath_rules):
        results = self.gets(tools.to_list(xpath_rules))
        return results[0] if results else None

    def gets(self, xpath_rules):
        if self.etree is None:
            return []

        for xpath_rule in tools.to_list(xpath_rules):
            _result = self.etree.xpath(xpath_rule)
            if _result:
                return _result

        return []

def _add_proxies_param(kwargs):
    proxy_node = _proxy_manager.get_best_proxy_node()
    if proxy_node:
        proxy_url = 'http://{username}:{password}@{ip}:{port}'.format(
                **proxy_node)
        kwargs['proxies'] = {
            'http'  :   proxy_url,
            'https' :   proxy_url,
            }

        kwargs.setdefault('timeout', (2, 40)) # 如果走得是代理，则把超时调小

    else:
        kwargs.pop('proxies', None)

    return kwargs

def _remove_proxies_param(kwargs):
    kwargs.pop('proxies', None)

    return kwargs

def _need_proxy_this_url(url, proxy_domain_rules):
    """根据url判断该url是否需要走代理"""
    if not proxy_domain_rules:
        return True

    for proxy_domain_rule in tools.to_list(proxy_domain_rules):
        if re.search(proxy_domain_rule, url):
            return True

    return False


def request(method, url, *args, **kwargs):
    """
    session: 指定使用哪个session, 不指定则表示使用requests
    retry: 重试多少次， 默认2次
    retry_interval: 重试间隔, 默认0
    proxy_domain: 只对部分域名做代理，写正则表达式即可, 可以传入string, tuple, list
    """
    session = kwargs.pop('session', requests)
    retry = kwargs.pop('retry', int(REQUEST_RETRY))
    proxy_domain = kwargs.pop('proxy_domain', None)

    retry_interval = kwargs.pop('retry_interval', int(REQUEST_RETRY_INTERVAL))

    if _need_proxy_this_url(url, proxy_domain):
        _add_proxies_param(kwargs)

    _ex_counter = 0
    _proxy_ex_counter = 0
    response = None
    proxy_has_error = False
    while True:
        try:
            response = session.request(method, url, *args, **kwargs)
            _proxy_manager.incr_use_counter()

        except requests.exceptions.RequestException as ex:
            if 'ProxyError' in str(ex.message):
                if _proxy_ex_counter < int(PROXY_RETRY):
                    if 'timed out' in str(ex.message): # 如果是超时了，就不理，表示已经有重拨过了
                        print('__________', str(ex.message), '_________')
                        pass

                    else:
                        _proxy_manager.flush_cache()

                    _proxy_ex_counter += 1
                    _add_proxies_param(kwargs)

                else:
                    _remove_proxies_param(kwargs)

                continue


            _ex_counter += 1
            if _ex_counter > retry:
                raise

            retry_interval and time.sleep(retry_interval)
            continue

        break

    response.selector = HTMLSelector(response)

    return response

get = partial(request, 'GET')
post = partial(request, 'POST')
delete = partial(request, 'DELETE')
put = partial(request, 'PUT')

import unittest
class TestCase(unittest.TestCase):
    @mock.patch.object(_proxy_manager, 'get_best_proxy_node')
    def test_has_proxy_node(self, mock_get_best_proxy_node):
        mock_get_best_proxy_node.return_value = {
                'ip'            :           '123.123.123.123',
                'port'          :           '1234',
                'username'      :           'ljd',
                'password'      :           'dsb',
                }

        kwargs = {}
        _add_proxies_param(kwargs)

        self.assertIn('proxies', kwargs)
        self.assertEqual(kwargs['proxies']['http'],
                'http://ljd:dsb@123.123.123.123:1234')

    @mock.patch.object(_proxy_manager, 'get_best_proxy_node')
    def test_no_proxy_node(self, mock_get_best_proxy_node):
        mock_get_best_proxy_node.return_value = None
        kwargs = {}
        _add_proxies_param(kwargs)
        self.assertNotIn('proxies', kwargs)

    def test_need_proxy_this_url(self):
        self.assertFalse(_need_proxy_this_url('http://www.baidu.com', r'.+\.taobao.com'))
        self.assertTrue(_need_proxy_this_url('http://www.taobao.com', r'.+\.taobao.com'))




#!/usr/bin/env python
#coding:utf-8
# Author        : tuxpy
# Email         : q8886888@qq.com.com
# Last modified : 2016-03-17 11:09:25
# Filename      : __init__.py
# Description   : 
from __future__ import print_function, unicode_literals

from ._requests import (request, get, post, delete, put, flush_ip)
import requests

session = requests.session

